<div align="center"><h1><b><font size=7>Proj6-实现一个基于eBPF技术监控Linux系统稳定性的工具</font></b></h1></div>
<div align="center"><font size=3>2024 年全国大学生计算机系统能力大赛操作系统设计赛——功能赛道</font></div>
<div align="center"><font size=5>队伍名称：ZAIN</font></div>
<div align="center"><font size=5>队员姓名：李瑞东</font></div>
<div align="center"><font size=5>指导教师：舒新峰教授</font></div>
<div align="center"><font size=5>学校名称：西安邮电大学</font></div>


### 项目名称：

基于 eBPF 的 Linux 系统稳定性观测工具

### 1.项目描述：

本项目通过 eBPF 技术开发一套专注于 Linux 系统稳定性的实时观测工具，从 CPU、内存、IO 和网络四个关键领域出发，捕获内核运行时的状态与行为数据，实现不稳定性检测、故障归因分析和优化建议生成。工具以低开销、高实时性和可扩展性为特点，为复杂场景下的性能监测和故障诊断提供保障，有效提升系统稳定性，助力云计算、大数据等核心业务的可靠运行。

### 2.项目结构
```
.
├── CPU_Subsystem
│   ├── blazesym
│   ├── bpftool
│   ├── cpu_watcher
│   ├── libbpf
│   ├── vmlinux
│   └── vmlinux.h
├── FS_Subsystem
│   ├── blazesym
│   ├── bpftool
│   ├── fs_watcher
│   ├── libbpf
│   ├── LICENSE
│   ├── README.md
│   ├── tools
│   └── vmlinux.h
├── fs_watcher
│   ├── projects
│   └── rust
├── images
│   ├── cpu -c.png
│   ├── cpu -S.png
│   ├── fs.png
│   ├── mem -a.png
│   ├── mem -s.png
│   ├── net -i.png
│   └── net -t.png
├── Memory_Subsystem
│   ├── blazesym
│   ├── docs
│   ├── mem_watcher
│   └── vmlinux
├── Network_Subsystem
│   ├── libbpf-bootstrap
│   └── net_watcher
├── README1.md
└── 阶段性技术报告.md
```

### 3.主要工作：

1. 深入调研 Linux 系统稳定性监控需求，分析传统观测技术的局限性，明确基于 eBPF 技术开发 Linux 内核观测工具的必要性和优势。
2. 选定 eBPF 技术作为项目的核心，规划开发涵盖 CPU、内存、IO、网络四个方面，不少于 12 个性能观测点的 Linux 内核观测工具。
3. 详细设计各性能观测点，包括 CPU 观测点（如内核函数 `schedule()` 执行时长、系统调用延迟）、内存观测点（如页面分配、内存回收）、IO 观测点（如文件写入操作）、网络观测点（如 HTTP 信息监控、数据包处理时延），并阐述其对系统稳定性分析的重要性。
4. 开发多个“watcher”工具，包括 `cpu_watcher`、`mem_watcher`、`fs_watcher`、`net_watcher`，实现对 Linux 系统性能的全面监测。
5. 实现 CPU 监测工具“cpu_watcher”，通过 eBPF 技术实时监测内核函数 `schedule()` 的执行时长和系统调用的延迟时间，精准检测 CPU 性能问题。
6. 开发内存监测工具“mem_watcher”，跟踪内核中页面的回收行为，提取内存相关统计数据，帮助分析内存使用情况和内存回收机制的效率。构建 IO 监测工具“fs_watcher”，监控文件系统中 `vfs_write` 系统调用的相关行为，收集文件写入操作的关键信息，辅助进行系统性能分析和故障排查。
7. 打造网络监测工具“net_watcher”，基于 eBPF 高效收集网络协议栈信息，监控 HTTP 传输协议信息以及 TCP/UDP/ICMP 数据包在不同层的处理时延，深入分析网络性能。
8. 进行功能测试，验证各“watcher”工具的监测功能，确保其能够准确输出性能指标数据。

 ### 4.观测点

| 类别 | 观测点名称 | 功能描述 |
| --- | --- | --- |
| CPU | `SEC("kprobe/schedule")` | 当内核函数 schedule() 即将被调用时触发，记录当前时间戳，作为计算函数执行时长的起始时间点，捕捉进程调度操作开始的瞬间。 |
| CPU | `SEC("kretprobe/schedule")` | 在 schedule() 函数执行完毕并即将返回时激活，获取此时的时间戳，结合之前记录的起始时间，计算函数的执行时长，测量进程调度所消耗的时间。 |
| CPU | `SEC("tracepoint/raw_syscalls/sys_enter")` | 当系统调用即将执行时触发，记录当前时间戳，作为系统调用开始时间，把握系统调用的发起时刻，为计算系统调用延迟提供起始依据。 |
| CPU | `SEC("tracepoint/raw_syscalls/sys_exit")` | 在系统调用执行完毕并即将返回时激活，获取此时的时间戳，结合之前记录的开始时间，计算系统调用的延迟时间，测量每个系统调用的执行耗时。 |
| 内存 | `SEC("kprobe/mm/page_alloc")` | 当内核执行 get_page_from_freelist 函数以从空闲页面列表获取页面时触发，获取内存分配相关的关键参数，如内存分配标志（gfp_mask）、页面分配顺序（order）以及分配上下文（ac）等信息，为分析内存分配行为提供基础数据。 |
| IO | `kprobe/vfs_write` | 聚焦于内核文件写入操作，捕捉进程身份、目标文件标识以及写入操作的预期与实际数据量等关键要素，探究文件写入流程的实际表现，排查写入异常、分析进程与文件的交互模式以及文件系统的写入性能与健康状况。 |
| 网络 | `kprobe/tcp_v4_rcv` | 跟踪接收 IPv4 的 TCP 数据包，捕获进入 TCP 层的数据包，获取 struct sk_buff 结构体中的数据包相关信息，检查数据包内容判断是否包含 HTTP 请求头。 |
| 网络 | `kprobe/tcp_v6_rcv` | 跟踪接收 IPv6 的 TCP 数据包，捕获进入 TCP 层的数据包，获取 struct sk_buff 结构体中的数据包相关信息，检查数据包内容判断是否包含 HTTP 请求头。 |
| 网络 | `kprobe/tcp_sendmsg` | 跟踪 TCP 发送包信息，在发送 HTTP 响应时触发，获取数据包进入 TCP 层时刻的时间戳等信息，记录响应相关信息。 |
| 网络 | `kprobe/eth_type_trans` | 进入 eth_type_trans 函数的入口，处理链路层协议的关键位置，根据 protocol_count 判断是否进行协议统计，记录数据包在链路层的处理起始时间。 |
| 网络 | `kprobe/ip_rcv_core`（IPv4） | 跟踪记录 IPv4 数据包在内核中的处理时间，记录数据包在网络层（IP 层）的处理时间，计算从链路层进入网络层后的时延。 |
| 网络 | `kprobe/ip6_rcv_core`（IPv6） | 跟踪记录 IPv6 数据包在内核中的处理时间，记录数据包在网络层（IP 层）的处理时间，计算从链路层进入网络层后的时延。 |
| 网络 | `kprobe/tcp_v4_do_rcv` | 获取套接字 sk 和其他信息，与接收函数配合处理数据包在传输层（TCP 层）的信息，计算数据包从网络层进入传输层后的时延。 |
| 网络 | `kprobe/tcp_v6_do_rcv` | 获取套接字 sk 和其他信息，与接收函数配合处理数据包在传输层（TCP 层）的信息，计算数据包从网络层进入传输层后的时延。 |
| 网络 | `kprobe/skb_copy_datagram_iter` | 处理网络数据包，记录分析包在不同网络层之间的时间差，分 IPv4 以及 IPv6 情况，计算传输层内部处理的时延。 |
| 网络 | `kprobe/ip_queue_xmit`（IPv4） | 获取 IPv4 数据包进入 IP 层时刻的时间戳，记录数据包从传输层进入网络层（IP 层）后的时延。 |
| 网络 | `kprobe/inet6_csk_xmit`（IPv6） | 获取 IPv6 数据包进入 IP 层时刻的时间戳，记录数据包从传输层进入网络层（IP 层）后的时延。 |
| 网络 | `kprobe/__dev_queue_xmit` | 获取数据包进入数据链路层时刻的时间戳，计算数据包从网络层进入链路层后的时延。 |
| 网络 | `kprobe/dev_hard_start_xmit` | 获取数据包发送时刻的时间戳，计算最终发送的时延，得到数据包在整个发送过程中经过不同层的时延。 |
| 网络 | `kprobe/icmp_rcv` | 用于接收 ICMP 数据包，记录时间戳，计算 ICMP 数据包在链路层、网络层等不同层的处理时延。 |
| 网络 | `kprobe/__sock_queue_rcv_skb` | 用于在接收 ICMP 数据包后的一些处理操作，记录时间戳，计算 ICMP 数据包在不同层的处理时延。 |
| 网络 | `kprobe/icmp_reply` | 用于发送 ICMP 回复，记录时间戳，计算 ICMP 数据包在不同层的处理时延。 |

### 5.比赛视频展示

通过百度网盘分享的文件：Ubuntu项目 - VMware Workstation 2024-...
链接：https://pan.baidu.com/s/1KpvcSvYKqkdEwsn7mQNlwg?pwd=no43 
提取码：no43


### 6.项目进展：

| 任务编号 | 任务描述                                                     | 完成情况 | 完成细节                                                     |
| -------- | ------------------------------------------------------------ | -------- | ------------------------------------------------------------ |
| 任务1    | 使用 eBPF 技术开发一套Linux内核观测工具，包含CPU、内存、IO、网络四个方面，这些方面可以精准检测到系统即将产生不稳定问题，Linux内核观测工具不少于12个性能观测点。 | 100%完成 | 设计并实现了12个以上的性能观测点。 利用eBPF技术，确保了工具的安全性和低开销。 |
| 任务2    | 精确度量用户的各类稳定性问题并能有效抓取现场并完成归因分析。 | 100%完成 | 实现了对系统稳定性问题的实时监控和度量。 工具能够捕获现场数据，为问题分析提供了详实的信息。 |
| 任务3    | 及时调整系统资源或者调度行为来减少直到消除各类稳定性问题。   | 100%完成 | 根据监控信息，我们可以做出相对的                             |         