// #include <linux/bpf.h>
#include "vmlinux.h"
#include <bpf/bpf_helpers.h>
#include <bpf/bpf_tracing.h>
#include <bpf/bpf_core_read.h>
#include "write.h"

char LICENSE[] SEC("license") = "Dual BSD/GPL";

struct
{
	__uint(type, BPF_MAP_TYPE_RINGBUF);
	__uint(max_entries, 256 * 1024);
} rb SEC(".maps");



SEC("kprobe/vfs_write")
int kprobe_vfs_write(struct pt_regs *ctx)
{
  pid_t pid;
  unsigned long inode_number;//定义用于存储inode号码的变量

   //探测的是第一个参数，文件指针,读取inode_number
  struct file *filp = (struct file *)PT_REGS_PARM1(ctx); 
  struct dentry *dentry = BPF_CORE_READ(filp,f_path.dentry);
  struct inode *inode = BPF_CORE_READ(dentry,d_inode);
  int ret = bpf_probe_read_kernel(&inode_number,sizeof(inode_number),&inode->i_ino);

  //探测的是第三个参数，要写入的字节数
  size_t count = (size_t)PT_REGS_PARM3(ctx);
  
  //这是vfs_write的返回值，它是一个实际写入的字节数
  size_t real_count = PT_REGS_RC(ctx);
  
  pid = bpf_get_current_pid_tgid() >> 32;
 
    

	struct sys_call_info *sc_if;
  
	sc_if = bpf_ringbuf_reserve(&rb, sizeof(*sc_if), 0);
  // bpf_printk("111");
	if (!sc_if)
		return 0;
  
  sc_if->pid = pid;
  sc_if->real_count = real_count;
  sc_if->count = count;
  sc_if->inode_number = inode_number;

  bpf_ringbuf_submit(sc_if, 0);
  return 0;
}   

