<div align="center"><h1><b><font size=7>Proj6-实现一个基于eBPF技术监控Linux系统稳定性的工具</font></b></h1></div>
<div align="center"><font size=3>2024 年全国大学生计算机系统能力大赛操作系统设计赛——功能赛道</font></div>
<div align="center"><font size=5>队伍名称：ZAIN</font></div>
<div align="center"><font size=5>队员姓名：李瑞东</font></div>
<div align="center"><font size=5>指导教师：舒新峰教授</font></div>
<div align="center"><font size=5>学校名称：西安邮电大学</font></div>



# 一、文档概述

Linux 系统在服务器、云计算、物联网和超级计算机等领域广泛应用，其稳定性对应用至关重要。传统观测技术依赖内核暴露数据，存在局限且内核改动受限。eBPF 技术源于 Linux 内核，能在特权上下文运行沙盒程序，安全扩展内核功能，具有细粒度、观测点广、可编程、低开销等特性，是监控 Linux 系统稳定性的理想选择。本项目将基于 eBPF 技术开发涵盖 CPU、内存、IO、网络四个方面，不少于 12 个性能观测点的 Linux 内核观测工具，实现系统不稳定问题的检测、归因分析与解决策略，以满足大赛要求。

# 二、技术背景与目标

### 1.Linux 系统稳定性的重要性

Linux 在多领域广泛应用，是众多关键业务的支撑。服务器领域承载企业关键应用，云计算提供弹性资源，物联网用于智能设备，超级计算机助力科学计算。其稳定性影响业务连续性，传统观测技术依赖内核暴露数据，存在局限，内核改动困难，无法满足稳定性监控需求。

### 2.eBPF 技术优势

eBPF 起源于 Linux 内核，可在特权上下文安全运行沙盒程序，扩展内核功能而无需修改源代码或加载模块。细粒度观测深入系统微观层面，广泛观测点覆盖系统各环节，可编程性支持定制逻辑，低开销确保对系统性能影响小，成为监控系统稳定性的理想技术。

### 3.项目目标

开发基于 eBPF 的 Linux 内核观测工具，精准检测系统不稳定问题，通过数据分析归因问题根源，依此提供解决策略，保障系统稳定运行，满足大赛对操作系统功能挑战的要求，提升 Linux 系统稳定性保障能力。

# 三、观测点说明

### （一）CPU观测点：

- **`SEC("kprobe/schedule")`**：当内核函数 schedule () 即将被调用时，该观测点被触发。在此处，程序可以记录当前时间戳，作为计算函数执行时长的起始时间点。这使得我们能够精确捕捉到进程调度操作开始的瞬间，为后续准确测量调度函数的执行时间奠定基础。
- **`SEC("kretprobe/schedule")`**：在 schedule () 函数执行完毕并即将返回时，此观测点被激活。程序获取此时的时间戳，并结合之前在`SEC("kprobe/schedule")`处记录的起始时间，计算出函数的执行时长。通过这两个观测点的协同工作，我们可以精确测量每次进程调度所消耗的时间，从而深入分析进程调度的效率和开销。

- **`SEC("tracepoint/raw_syscalls/sys_enter")`**：当系统调用即将执行时，该观测点触发。在此处，程序记录当前时间戳，作为系统调用开始时间。这有助于我们准确把握系统调用的发起时刻，为后续计算系统调用延迟提供起始依据。
- **`SEC("tracepoint/raw_syscalls/sys_exit")`**：在系统调用执行完毕并即将返回时，此观测点被激活。程序获取此时的时间戳，并结合之前记录的开始时间，计算出系统调用的延迟时间。通过这两个观测点的配合，我们可以精确测量每个系统调用的执行耗时，了解系统调用对 CPU 资源的占用情况以及系统处理系统调用的效率。

### （二）内存观测点：

- **观测点**：`SEC("kprobe/mm/page_alloc")`：当内核执行`get_page_from_freelist`函数以从空闲页面列表获取页面时，该观测点被触发。在此处，程序能够获取与内存分配相关的关键参数，如内存分配标志（`gfp_mask`）、页面分配顺序（`order`）以及分配上下文（`ac`）等信息。这些信息为后续深入分析内存分配行为提供了基础数据。

### （三）IO 观测点：

- 文件写入观测点：

`kprobe/vfs_write` ：观测点聚焦于内核文件写入操作。这一观测点犹如系统文件写入活动的探测器，能够捕捉到诸如进程身份、目标文件标识以及写入操作的预期与实际数据量等关键要素。它为深入探究文件写入流程在系统运行时的实际表现提供了窗口，有助于排查写入异常、分析进程与文件的交互模式以及整体文件系统的写入性能与健康状况等多方面的系统级行为与状态。

### （四）网络观测工具

1.**功能描述**

- 本模块专注于网络性能的观测，通过多个特定的观测指标来深入了解网络的运行状态，为系统稳定性分析提供关键数据支持。

1. HTTP 信息监控相关观测点
   - `kprobe/tcp_v4_rcv`和`kprobe/tcp_v6_rcv`：这两个观测点分别用于跟踪接收 IPv4 和 IPv6 的 TCP 数据包。当 HTTP 请求通过 TCP 传输时，这些观测点可以捕获进入 TCP 层的数据包。在函数`BPF_KPROBE(tcp_v4_rcv, struct sk_buff *skb)`和`BPF_KPROBE(tcp_v6_rcv, struct sk_buff *skb)`中，可以获取到`struct sk_buff`结构体，其中包含了数据包的相关信息。对于 HTTP 监控来说，可以通过检查数据包内容来判断是否包含 HTTP 请求头。因为 HTTP 通常是基于 TCP 协议传输的，所以接收 TCP 数据包是获取 HTTP 请求头的第一步。
   - `kprobe/tcp_sendmsg`这个观测点用于跟踪 TCP 发送包信息。在发送 HTTP 响应时，会触发这个观测点。通过这个观测点可以获取数据包进入 TCP 层时刻的时间戳等信息。在函数`BPF_KPROBE(tcp_sendmsg, struct sock *sk, struct msghdr *msg, size_t size)`中，可以获取到套接字`sk`、消息头`msg`和消息大小`size`等信息。对于 HTTP 监控，当服务器发送 HTTP 响应时，这个观测点可以帮助记录响应相关信息。
2. 监控 TCP/UDP/ICMP 数据包在不同层的处理时延相关观测点
   - 接收路径相关观测点
     - `kprobe/eth_type_trans`：这个观测点是进入`eth_type_trans`函数的入口。对于接收数据包来说，它是处理链路层协议的一个关键位置。在这里可以根据`protocol_count`判断是否进行协议统计（`sum_protocol(skb, false)`用于接收情况）。它是数据包进入网络协议栈后比较早期的处理位置，对于记录数据包在链路层的处理起始时间很重要。对于 IPv4 和 IPv6 的数据包接收都适用。
     - `kprobe/ip_rcv_core`（IPv4）和`kprobe/ip6_rcv_core`（IPv6）：这两个观测点分别用于跟踪记录 IPv4 和 IPv6 数据包在内核中的处理时间。`kprobe/ip_rcv_core`对应的函数`BPF_KPROBE(ip_rcv_core, struct sk_buff *skb)`用于处理 IPv4 数据包，`kprobe/ip6_rcv_core`对应的函数`BPF_KPROBE(ip6_rcv_core, struct sk_buff *skb)`用于处理 IPv6 数据包。它们可以记录数据包在网络层（IP 层）的处理时间，通过在这些函数中记录时间戳，就可以计算数据包在从链路层进入网络层后的时延。
     - `kprobe/tcp_v4_rcv`（IPv4）和`kprobe/tcp_v6_rcv`（IPv6）、`kprobe/tcp_v4_do_rcv`、`kprobe/tcp_v6_do_rcv`和`kprobe/skb_copy_datagram_iter`：kprobe/tcp_v4_rcv`和`kprobe/tcp_v6_rcv`用于接收 TCP 数据包，分别针对 IPv4 和 IPv6 协议，记录数据包在 TCPv4 和 TCPv6 层的时间戳。`kprobe/tcp_v4_do_rcv`和`kprobe/tcp_v6_do_rcv`用于获取套接字`sk`和其他信息，与前面的接收函数配合可以进一步处理数据包在传输层（TCP 层）的信息。`kprobe/skb_copy_datagram_iter`用于处理网络数据包，记录分析包在不同网络层之间的时间差，分 IPv4 以及 IPv6 情况。这些观测点可以帮助计算数据包从网络层进入传输层后的时延，以及在传输层内部处理的时延。
   - 发送路径相关观测点
     - `kprobe/tcp_sendmsg`：如前面所述，它用于获取数据包进入 TCP 层时刻的时间戳，是发送 tcp 层起始点。对于记录数据包在传输层的发送起始时间很重要，可以与接收路径中传输层的时间戳对比计算时延。
     - `kprobe/ip_queue_xmit`（IPv4）和`kprobe/inet6_csk_xmit`（IPv6）：这两个观测点分别用于获取 IPv4 和 IPv6 数据包进入 IP 层时刻的时间戳。在函数`BPF_KPROBE(ip_queue_xmit, struct sock *sk, struct sk_buff *skb)`和`BPF_KPROBE(inet6_csk_xmit, struct sock *sk, struct sk_buff *skb)`中可以获取相关信息。通过记录这些时间戳，可以计算数据包从传输层进入网络层（IP 层）后的时延。
     - `kprobe/__dev_queue_xmit`和`kprobe/dev_hard_start_xmit`：`kprobe/__dev_queue_xmit`用于获取数据包进入数据链路层时刻的时间戳，`kprobe/dev_hard_start_xmit`用于获取数据包发送时刻的时间戳。通过在这两个观测点记录时间戳，结合前面各层的时间戳，就可以计算数据包在从网络层进入链路层后的时延，以及最终发送的时延，从而得到数据包在整个发送过程中经过不同层的时延。
   - ICMP 相关观测点
     - `kprobe/icmp_rcv`、`kprobe/__sock_queue_rcv_skb`和`kprobe/icmp_reply`：`kprobe/icmp_rcv`用于接收 ICMP 数据包，`kprobe/__sock_queue_rcv_skb`用于在接收 ICMP 数据包后的一些处理操作，`kprobe/icmp_reply`用于发送 ICMP 回复。虽然 ICMP 不是基于 TCP 或 UDP 的传输协议，但对于完整的网络数据包时延监控来说，也需要考虑 ICMP 数据包在不同层的处理时延。通过在这些观测点记录时间戳，可以计算 ICMP 数据包在链路层、网络层等不同层的处理时延，计算方式与 TCP/UDP 数据包类似。

# 四、设计思路与工具说明

开发多个"watcher"工具，包括：

1. `cpu_watcher`：监测CPU性能指标。
2. `mem_watcher`：监测内存使用情况。
3. `net_watcher`：监测网络流量及连接状态。
4. `fs_watcher`：监测文件系统的IO性能。

### （一）CPU观测 工具：

1. **功能描述**
   - 本模块专注于内存性能的观测，通过两个特定的观测指标来深入了解内存的运行状态，为系统稳定性分析提供关键数据支持。
   - **详细说明**

   - cs_delay（实时对内核函数 schedule () 的执行时长进行测试）

   - **工具说明**：
     - Cs_delay是利用eBPF技术对内核函数schedule()的执行时长进行测试的工具，该工具使用了BPF环形缓冲区（ringbuf），避免了BPF性能缓冲区（perfbuf）导致的内存使用效率低和事件重新排序等问题。


- syscall_delay（实时采集当前系统调用时间）

  - **工具说明**：

    - syscall_delay 是一款专门用于捕获进程系统调用行为的工具，旨在识别系统中每个线程和进程的系统调用活动。通过在 sys_enter 和 sys_exit 这两个观测点进行监控，该工具能够检测到所有产生的系统调用，并计算每次系统调用的延迟时间，从而实现对系统中系统调用行为的宏观分析。工具可以采集到的性能指标如表3-1所示：

    | **性能指标**   | **含义**             |
    | -------------- | -------------------- |
    | **Time**       | 发生系统调用的时间点 |
    | **Pid**        | 发生系统调用的进程号 |
    | **Syscall_id** | 触发的系统调用号     |
    | **delay/ms**   | 该系统调用的延迟     |

​																				表3.1 syscall_delay性能指标

### （二）内存观测工具：

1. **功能描述**

   - 本模块专注于内存性能的观测，通过一个特定的观测指标来深入了解 CPU 的运行状态，为系统稳定性分析提供关键数据支持。

2. **详细说明**

   - paf （实时检测 get_page_from_freelist(),获取页面）

     - **工具说明**：跟踪内核中页面的回收行为，记录回收的各个阶段，例如要回收的页面，以回收的页面，等待回收的脏页数，要写回的页数(包括交换空间中的页数)以及当前正在写回的页数。可以帮助监控系统内存的使用情况，预测内存不足的可能性，以及在必要时触发相应的内存回收机制来保证系统的稳定性和性能。

       表3.2中的paf性能指标记录了内存管理区各水位的页面数量以及内存回收相关的状态信息，帮助监控和调节内存回收机制，以保障系统内存的稳定性和性能表现。

       | **指标名称** | **指标含义**                         |
       | ------------ | ------------------------------------ |
       | min          | 内存管理区处于最低警戒水位的页面数量 |
       | low          | 内存管理区处于低水位的页面数量       |
       | high         | 内存管理区处于高水位的页面数量       |
       | present      | 内存管理区实际管理的页面数量         |
       | flag         | 申请页面时的权限（标志）             |

​																						 表3.2 paf性能指标	

- sysstat（提取内存相关统计数据）

  - **工具说明**：sysstat提取各种类型内存的活动和非活动页面数量，以及其他内存回收相关的统计数据，除了常规的事件信息外，还输出了与内存管理相关的详细信息。包括活跃内存、不活跃内存、不同类型内存的大小等信息，从而进行有效的资源管理和性能优化。

    表3.3中的sysstat性能指标总结了系统内存的活跃与非活跃状态，以及各类内存管理细节，有助于监测内存使用情况，优化内存分配，提升系统性能和资源利用效率。

    | **指标名称**   | **指标含义**                     |
    | -------------- | -------------------------------- |
    | active         | LRU活跃内存大小                  |
    | inactive       | LRU不活跃内存大小                |
    | anon_active    | 活跃匿名内存大小                 |
    | anon_inactive  | 不活跃匿名内存大小               |
    | file_active    | 活跃文件映射内存大小             |
    | file_inactive  | 不活跃文件映射内存大小           |
    | unevictable    | 不可回收内存大小                 |
    | dirty          | 脏页大小                         |
    | writeback      | 正在回写的内存大小               |
    | anonpages      | RMAP页面                         |
    | mapped         | 所有映射到用户地址空间的内存大小 |
    | shmem          | 共享内存                         |
    | kreclaimable   | 内核可回收内存                   |
    | slab           | 用于slab的内存大小               |
    | sreclaimable   | 可回收slab内存                   |
    | sunreclaim     | 不可回收slab内存                 |
    | NFS_unstable   | NFS中还没写到磁盘中的内存        |
    | writebacktmp   | 回写所使用的临时缓存大小         |
    | anonhugepages  | 透明巨页大小                     |
    | shmemhugepages | shmem或tmpfs使用的透明巨页       |

    ​																	表3.3 sysstat性能指标

    通过监测这些指标，可以发现潜在的内存泄漏问题，优化内存分配策略，合理分配系统资源，确保系统稳定性和数据一致性。此外，对于透明巨页和共享内存等特殊内存类型，提取相应指标有助于优化其管理和利用方式，提高系统整体的性能和效率。综合而言，这些指标的提取为系统管理和性能优化提供了重要的数据支持，有助于提升系统的可靠性和性能表现。

### （三）IO 观测工具：

1.**功能描述**

- 本模块专注于IO性能的观测，通过一个特定的观测指标来深入了解IO的运行状态，为系统稳定性分析提供关键数据支持。

2.**详细说明**

- write 信息监控：

**工具说明**`：`write`工具主要用于在内核态监控文件系统中`vfs_write`系统调用的相关行为，并收集关键信息传递到用户态，旨在帮助分析文件写入操作的相关情况，比如哪些进程进行了文件写入、写入的目标文件信息以及预期和实际的写入字节数等情况，辅助进行系统性能分析、故障排查等工作。

### （四）网络观测工具

1.**功能描述**

- 本模块专注于网络性能的观测，通过多个特定的观测指标来深入了解网络的运行状态，为系统稳定性分析提供关键数据支持。

2.**详细说明**

- **HTTP 信息监控**

​	**工具说明**：对HTTP传输协议信息进行监控，在指定-i参数后，从携带HTTP请求头的TCP包中将HTTP请求头提取并输出，提取信息包括包含请求资源方式、状态码、状态文本等。

- **监控 TCP/UDP/ICMP 数据包在不同层的处理时延**

记录数据包在收包和发包过程中经过不同层（链路层、网络层、传输层）的时延之差（单位微秒），支持 ipv4、ipv6 协议。

表 3.4中的性能指标记录了记录链路层、网络层和传输层的时延差异，以及各个处理链的时延（如 PRE_ROUTING、LOCAL_IN 等）相关状态信息，全面覆盖网络协议栈，精确分析数据包在不同协议层的延迟，帮助定位网络性能瓶颈和问题，确保网络流量在高负载下依旧能保持稳定运行。

| 指标名称  | 指标含义                |
| --------- | ----------------------- |
| sock      | Socket 地址             |
| Saddr     | 源地址                  |
| Daddr     | 目的地址                |
| Sport     | 源端口                  |
| Dport     | 目的端口                |
| rx/flag   | 收包 / 发包方向         |
| len       | 流量                    |
| time      | 时延                    |
| SEQ       | TCP 报文段序列号        |
| ACK       | TCP 报文段确认号        |
| MAC_TIME  | TCP 数据包链路层时延    |
| IP_TIME   | TCP 数据包网络层时延    |
| TRAN_TIME | TCP 数据包传输层时延    |
| PreRT     | PRE_ROUTING 链处理时延  |
| L_IN      | LOCAL_IN 链处理时延     |
| PostRT    | POST_ROUTING 链处理时延 |
| L_OUT     | LOCAL_OUT 链处理时延    |

​																		表 3.4 时延指标

# 五、测试

## （一）功能测试

#### 1.cpu_watcher工具测试

适配命令：

| **监测指标**       | **适配命令**          |
| ------------------ | --------------------- |
| 系统调用运行延迟   | sudo ./cpu_watcher -S |
| Schedule()执行时间 | sudo ./cpu_watcher -c |

运行结果：

- -c工具：

![cpu -c](<images/cpu -c.png>)

解释：

这张图展示了网络延迟（`cs_delay`）数据分布的相关信息，从左到右的指标依次是网络延迟值（`cs_delay`）、出现次数（`Count`）和分布情况（`Distribution`）。

从这张图可以看出，网络延迟在不同范围内的分布情况，有助于分析网络性能和定位可能存在的问题。例如，较低延迟（如 0 - 16）出现的频率较高，而较高延迟（如 262144 及以上）出现的频率较低。

- -S工具：

![cpu -S](<images/cpu -S.png>)

这张图展示了一些系统调用相关的数据，从左到右的指标依次是时间（Time）、进程 ID（Pid）、系统调用 ID（syscall_id）和延迟（delay/ms）。

从图中的数据可以看出，在 22:26:51 这个时间点，多个进程同时进行了不同类型的系统调用，并且每个系统调用的延迟时间各不相同。这些数据可以用于分析系统在特定时间点的性能，特别是不同进程的系统调用开销。

#### 2.mem_watcher工具测试

**适配命令**：

| **监测指标**             | **适配命令**          |
| ------------------------ | --------------------- |
| 跟踪内核中页面的回收行为 | sudo ./mem_watcher -a |
| 提取内存相关统计数据     | sudo ./cpu_watcher -S |

- -a工具：

![mem -a](<images/mem -a.png>)

这张图展示了一些与内存管理相关的数据，从左到右的指标依次是最小值（MIN）、低水位（LOW）、高水位（HIGH）、当前值（PRESENT）、标志（FLAG）和一些内存管理相关的操作（以下简称为操作列)。

从图中的数据可以看出，这些记录可能与内存管理相关，展示了在某个特定场景下，内存管理相关指标的固定值以及不同的标志和操作。这些数据可以用于分析系统在特定场景下的内存管理情况。

- -s工具：

![mem -s](<images/mem -s.png>)

这张图展示了与内存管理相关的数据，从左到右的指标依次是活跃内存（ACTIVE）、不活跃内存（INACTIVE）、非活动匿名内存（ANON_ACT）、非活动文件内存（ANON_INA）、文件内存（FILE_ACT）、文件非活动内存（FILE_INA）、未回收内存（UNEVCIT）、脏页（DIRTY）、回写（WRITEBK）、匿名页面（ANONPAG）、映射（MAP）和共享内存（SHMEM）。

从图中的数据可以看出，这些记录详细地展示了系统在不同时刻下各种类型内存的状态和数量，可用于分析内存管理的性能和状态，例如活跃内存和不活跃内存的比例、不同类型内存的使用情况等。

#### 3.fs_watcher 工具测试

**适配命令**：

| **监测指标** | **适配命令**         |
| ------------ | -------------------- |
|              | sudo ./fs_watcher -S |

![fs](images/fs.png)

#### 4.net_watcher 工具测试

**适配命令**：

| **监测指标**           | **适配命令**          |
| ---------------------- | --------------------- |
| 跟踪HTTP信息           | sudo  ./netwatcher -i |
| 跟踪各数据包每层的延迟 | sudo  ./netwatcher -t |

- -i 命令：

![net -i](<images/net -i.png>)

这张图展示了网络相关的数据，从左到右的指标依次是源 IP（SOCK）、TCP 报文段序列号（SEQ）、TCP 报文段确认号（ACK）、TCP 数据包链路层时延（MAC_TIME）、TCP 数据包网络层时延（IP_TIME）、TCP 数据包传输层时延（TRAN_TIME）和 HTTP 响应信息（HTTP）。

从图中的数据可以看出，这些记录展示了不同源 IP 的网络数据包在 TCP/IP 各层的时延情况，以及对应的 HTTP 请求和响应信息。这些数据可以用于分析网络性能和 HTTP 交互情况。

- -t 命令：

![net -t](<images/net -t.png>)

这张图展示了网络相关的数据，从左到右的指标依次是源 IP（SOCK）、TCP 报文段序列号（SEQ）、TCP 报文段确认号（ACK）、TCP 数据包链路层时延（MAC_TIME）、TCP 数据包网络层时延（IP_TIME）、TCP 数据包传输层时延（TRAN_TIME）和协议类型（HTTP）。

从图中的数据可以看出，这些记录展示了来自同一源 IP 地址的网络数据包在 TCP/IP 各层的时延情况。这些数据可以用于分析网络性能。

 

# 六、工具抓取和分析

## 6.1.CPU 相关

（一）精确度量 CPU 的各类稳定性问题并能有效抓取现场并完成归因分析

1. 利用 cs_delay 工具

- 性能监控：
  - cs_delay 工具通过 eBPF 技术对内核函数 schedule () 的执行时长进行测试。在 CPU 性能分析中，内核调度函数的执行时长是关键指标之一。通过实时监测 schedule () 的执行时长，可以了解 CPU 在内核调度层面的性能情况。例如，如果发现 schedule () 的执行时长突然变长，可能意味着内核调度出现了问题，如调度队列阻塞等。
- 问题抓取：
  - 当 CPU 出现性能异常时，cs_delay 所记录的内核函数调度时长数据能够帮助抓取现场。由于它使用了 BPF 环形缓冲区（ringbuf），能够更高效地记录数据，避免数据丢失和事件重新排序等问题。这些数据可以反映出在 CPU 性能下降时，内核调度是否存在异常情况。
- 归因分析：
  - 根据 cs_delay 记录的内核调度时长数据，可以分析 CPU 性能问题的原因。例如，如果发现某个时间段内 schedule () 执行时长持续增加，结合系统中运行的进程情况，可以进一步判断是由于某些进程的不合理资源占用（如 CPU 密集型进程过多）导致内核调度负担加重，还是内核调度算法本身在特定场景下出现了性能瓶颈。
  
  2. 利用 syscall_delay 工具
  
  - 性能监控：
    - syscall_delay 工具通过在 sys_enter 和 sys_exit 观测点进行监控，能够捕获系统中所有进程和线程的系统调用行为，并计算每次系统调用的延迟时间。对于 CPU 性能分析来说，系统调用的延迟情况会直接影响 CPU 的利用率和系统的响应速度。例如，频繁的高延迟系统调用可能导致 CPU 等待时间增加，从而影响整体性能。
  
  - 问题抓取：
    - 当 CPU 出现性能问题时，syscall_delay 所采集的系统调用延迟数据可以帮助确定是否是系统调用层面的问题导致。例如，如果发现某个进程的系统调用延迟突然大幅增加，可能意味着该进程在执行系统调用时遇到了阻塞情况，如等待 I/O 资源或锁资源等。
  

- 归因分析：
  - 根据 syscall_delay 记录的系统调用延迟数据和相关指标（如发生系统调用的时间点、进程号、系统调用号等），可以深入分析 CPU 性能问题的根源。例如，如果发现某个数据库进程的系统调用延迟较高，可能是数据库配置不合理导致频繁的系统调用等待，或者是与数据库相关的硬件资源（如存储设备）出现了性能瓶颈，进而影响了 CPU 的性能。

（二）及时调整系统资源或者调度行为来减少直到消除 CPU 各类稳定性问题

1. 利用 cs_delay 工具

- 资源调整：
  - 根据 cs_delay 所监测到的内核函数 schedule () 的执行时长，如果发现内核调度性能下降，可以考虑调整系统资源。例如，如果是由于多核心 CPU 的负载不均衡导致某些核心的调度任务过重，可以通过调整进程的亲和性，将部分进程重新分配到负载较轻的核心上，优化内核调度的性能。
- 调度优化：
  - 基于 cs_delay 的数据，可以对内核调度算法进行优化。如果发现 schedule () 在某些特定场景下执行时长过长，可以深入研究内核调度策略，考虑采用更适合当前系统负载情况的调度算法，或者对现有调度算法的参数进行调整，如调整时间片长度等，以提高 CPU 的调度效率。

2. 利用 syscall_delay 工具

- 资源调整：
  - 根据 syscall_delay 所采集的系统调用延迟数据，如果发现某些进程的系统调用延迟过高，可以调整系统资源来解决问题。例如，如果是由于 I/O 密集型进程的系统调用等待时间长，可以考虑增加 I/O 设备的带宽或者优化 I/O 调度策略，减少进程在系统调用时的等待时间，从而释放 CPU 资源，提高 CPU 的利用率。
- 调度优化：
  - 通过分析 syscall_delay 的数据，可以优化系统的调度行为。例如，如果发现某些系统调用频繁且延迟高，可以考虑对这些系统调用进行合并或者异步处理，减少 CPU 在等待系统调用完成时的空闲时间。同时，对于具有高优先级的系统调用，可以通过调整内核调度策略，确保这些调用能够及时得到处理，提高系统的整体性能和稳定性。

## 6.2内存相关

（一）精确度量内存的各类稳定性问题并能有效抓取现场并完成归因分析

1. 利用 paf 工具

- 性能监控：
  - paf 工具通过跟踪内核中页面的回收行为，能实时检测 get_page_from_freelist () 获取页面的情况。例如，它可以记录要回收的页面、已回收的页面、等待回收的脏页数、要写回的页数以及当前正在写回的页数等信息。这些数据有助于监控系统内存的使用情况，比如可以观察到内存回收的频率和规模，提前预测内存不足的可能性。
- 问题抓取：
  - 当系统出现内存相关的性能问题时，paf 记录的内存回收各阶段的数据可以帮助确定是否是内存回收机制出现异常。例如，如果发现等待回收的脏页数持续增加，可能意味着内存回收过程在脏页处理环节出现了阻塞。
- 归因分析：
  - 根据 paf 的性能指标（如内存管理区各水位的页面数量以及内存回收相关的状态信息），可以分析内存问题的原因。例如，如果发现内存管理区的低水位（low）页面数量快速下降接近最低警戒水位（min），结合系统中运行的进程对内存的需求情况，可以判断是某些进程过度占用内存导致内存紧张，还是内存回收机制未能及时释放足够的内存。

2. 利用 sysstat 工具

- 性能监控：
  - sysstat 工具能够提取各种类型内存的活动和非活跃页面数量，以及其他内存回收相关的统计数据。它可以提供诸如活跃内存、不活跃内存、不同类型内存的大小等详细信息。通过持续监测这些数据，可以了解内存的使用模式，例如是否存在大量内存长时间处于不活跃状态。
- 问题抓取：
  - 当内存出现问题时，sysstat 所提取的丰富内存数据可以帮助确定问题所在。例如，如果发现不可回收内存（unevictable）大小异常增加，可能暗示存在某些进程持有不可回收的内存资源，导致内存资源浪费。
- 归因分析：
  - 根据 sysstat 记录的各项内存指标，可以深入分析内存性能问题的根源。例如，如果发现活跃匿名内存（anon_active）和不活跃匿名内存（anon_inactive）的比例出现异常变化，结合系统中运行的应用程序特点，可以判断是某些应用程序的内存管理不当导致匿名内存使用异常，进而影响整个系统的内存性能。

（二）及时调整系统资源或者调度行为来减少直到消除内存各类稳定性问题

1. 利用 paf 工具

- 资源调整：
  - 根据 paf 所监测到的内存回收情况，如果发现等待回收的脏页数过多，可能需要调整系统资源来加速脏页的处理。例如，可以适当增加用于脏页写回的 I/O 带宽，或者优化脏页写回的调度策略，确保内存回收过程能够顺利进行，避免内存资源的过度占用。
- 调度优化：
  - 基于 paf 的数据，如果发现内存回收在某些阶段出现延迟或阻塞，可以对内存回收的调度进行优化。例如，调整页面回收的优先级策略，确保更重要的页面能够优先得到回收处理，提高内存回收机制的整体效率，保障系统内存的稳定性。

2. 利用 sysstat 工具

- 资源调整：
  - 根据 sysstat 所采集的内存数据，如果发现不活跃内存（inactive）数量较大，可以考虑调整内存分配策略。例如，将部分不活跃内存重新分配给活跃进程，提高内存的利用率。或者对于共享内存（shmem），如果发现其使用量过大且存在浪费现象，可以调整共享内存的分配和管理策略。
- 调度优化：
  - 通过分析 sysstat 的数据，如果发现某些内存类型（如 NFS_unstable）存在性能问题，可以优化相关的内存调度行为。例如，对于 NFS 相关的内存操作，可以优化网络文件系统的缓存策略，减少 NFS 中还没写到磁盘中的内存数量，提高内存操作的效率和系统的整体性能。

## 6.3 内存相关

（一）精确度量 IO 的各类稳定性问题并能有效抓取现场并完成归因分析

1. 利用 write 信息监控工具

- 性能监控：
  - write 信息监控工具通过捕获内核态 `vfs_write` 系统调用信息，可实时监测文件写入操作的频率与规模。例如，统计单位时间内不同进程发起的 `vfs_write` 调用次数，了解哪些进程对存储系统的写入需求较为频繁，从而评估系统整体的写入负载情况。同时，对比各个进程的预期写入字节数和实际写入字节数，能判断写入操作的有效性和效率。若发现某进程的实际写入字节数远低于预期，可能暗示存储设备存在写入性能问题或者文件系统层面的限制。
  - 借助收集到的目标文件 inode 号码信息，可以进一步分析写入操作在文件系统层面的分布情况。例如，确定哪些文件或目录是写入热点，是否存在大量写入集中在特定文件或目录导致局部存储压力过大的情况，这有助于提前发现潜在的 IO 瓶颈风险。
- 问题抓取：
  - 当系统出现 IO 性能问题时，write 信息监控工具记录的信息可精准定位异常。如果观察到大量进程的写入操作长时间处于等待状态（通过分析实际写入字节数长时间不更新），可能意味着存储设备繁忙或出现故障，如磁盘已满、磁盘 I/O 队列深度过大或者磁盘出现硬件故障导致写入延迟极高。
  - 若特定进程的写入操作频繁出现错误（可根据系统调用返回值判断），例如返回错误码表示权限不足或存储介质不可写等情况，可针对性地检查该进程的权限设置以及存储设备的挂载选项和状态，确定是配置错误还是设备故障导致。
- 归因分析：
  - 根据 write 信息监控工具提供的数据，结合系统的硬件配置和软件环境，可以深入剖析 IO 问题的根源。例如，如果发现多个进程的写入性能在某个时间段突然下降，且从收集到的信息中看到这些进程都在向同一物理磁盘的不同分区写入数据，那么可以考虑检查该磁盘的健康状况（如使用磁盘检测工具查看是否存在坏道）、磁盘的 I/O 调度策略是否合理以及系统内存对磁盘写入缓存的影响等因素。又如，如果发现某个应用程序的写入操作总是失败且错误码指向文件系统的配额限制，那么就需要检查文件系统针对该应用程序用户或用户组的配额设置是否合理，以及是否需要调整配额以满足业务需求。

（二）及时调整系统资源或者调度行为来减少直到消除 IO 各类稳定性问题

1. 利用 write 信息监控工具

- 资源调整：
  - 根据 write 信息监控工具监测到的写入操作情况，如果发现某个进程的写入需求过大，影响了其他关键业务的 IO 性能，可以对该进程的资源进行限制。例如，通过操作系统的资源管理工具，限制该进程能够占用的磁盘 I/O 带宽或每秒最大写入操作次数，确保关键业务有足够的 IO 资源。若发现写入操作集中在某个慢速存储设备上，可考虑将部分数据迁移到性能更好的存储介质（如从机械硬盘迁移到固态硬盘），或者增加存储设备的数量并采用磁盘阵列技术（如 RAID）来提高整体的写入性能和数据冗余性。
- 调度优化：
  - 基于 write 信息监控工具提供的信息，优化 IO 调度行为。例如，如果发现多个进程的写入操作存在竞争关系，导致整体写入效率低下，可以调整操作系统的 I/O 调度算法。对于对延迟敏感的写入操作（如数据库事务日志的写入），可以提高其调度优先级，确保其能够及时完成，减少对系统整体性能的影响。同时，根据写入操作在文件系统中的分布情况，合理调整文件系统的布局和缓存策略。例如，如果发现某个目录下的文件写入频繁，可以将该目录单独挂载到一个具有较大缓存空间的文件系统分区上，或者调整该分区的预读和写回策略，提高写入操作的效率。

## 6.4网络相关

（一）精确度量网络的各类稳定性问题并能有效抓取现场并完成归因分析

1. 利用监控数据包工具

- 性能监控：
  - 监控数据包工具可以实时抓取网络数据包，通过解析数据包中的信息，如源地址、目的地址、端口号、协议类型等，能够清晰地了解网络流量的流向与分布。例如，它可以统计出特定时间段内不同源地址发往不同目的地址的数据包数量，以及各种协议（TCP、UDP、ICMP 等）数据包所占的比例，从而全面掌握网络的负载情况以及流量构成。同时，还能监测数据包的大小分布，判断是否存在大量异常大小的数据包影响网络传输效率。
- 问题抓取：
  - 当网络出现性能问题时，所捕获的数据包能为确定问题提供关键线索。例如，如果发现某一源地址持续向多个目的地址发送大量数据包，且数据包的内容具有特定规律（如相同的请求信息），可能意味着存在恶意扫描或攻击行为。若检测到大量数据包的传输延迟过高，可进一步查看数据包的详细信息，确定是网络链路拥塞、网络设备故障还是特定应用程序的问题导致。例如，若某个特定端口的数据包延迟明显高于其他端口，可能是该端口对应的应用程序或服务出现异常。
- 归因分析：
  - 根据监控数据包工具提供的数据，结合网络拓扑结构和设备配置信息，可以深入分析网络问题的根源。例如，如果发现某一网段内的数据包丢包率较高，且从数据包信息中发现这些数据包都经过某一特定的网络交换机，那么可以重点排查该交换机的配置和运行状态，判断是交换机端口故障、配置错误（如 VLAN 划分不合理）还是负载过高导致数据包丢弃。又如，如果发现网络中特定类型的应用（如视频会议应用）数据包传输延迟增加，结合网络带宽占用情况以及其他相关设备（如路由器的 QoS 配置），可以确定是由于网络带宽不足、其他应用抢占资源还是网络设备对该类型数据包处理不当所引起的。

2. 利用 http 工具

- 性能监控：
  - http 工具能够对基于 HTTP 协议的网络交互进行监测与分析。它可以记录 HTTP 请求的频率、来源 IP、请求的 URL 路径以及响应状态码等信息。通过分析这些数据，能够了解哪些网站或服务的 HTTP 流量较大，以及这些流量是否正常。例如，统计特定时间段内对某个电商网站的 HTTP 请求数量，判断其业务流量的高峰时段和低谷时段，以便合理安排服务器资源。同时，根据响应状态码的分布情况，如大量的 500 错误码，可能表示服务器端存在内部错误，影响用户的访问体验。
- 问题抓取：
  - 当网络出现与 HTTP 相关的异常时，http 工具所记录的数据可帮助定位问题。例如，如果发现某个客户端的 IP 地址频繁发送大量的 HTTP 请求，且请求的 URL 路径异常（如不断尝试访问受限资源），可能存在恶意攻击或爬虫行为。若检测到大量的 HTTP 请求超时，可能是网络链路问题导致数据包传输缓慢，也可能是服务器负载过高，无法及时响应请求。例如，在电商促销活动期间，如果大量用户同时访问商品详情页面，http 工具监测到大量的请求延迟或失败，就需要进一步排查是网络带宽不足、服务器性能瓶颈还是应用程序代码问题导致的。
- 归因分析：
  - 根据 http 工具提供的 HTTP 相关数据以及网络环境信息，可以深入剖析网络问题的根源。例如，如果发现从某个地区的用户访问网站时经常出现 HTTP 连接中断，结合网络拓扑和该地区的网络服务提供商情况，可以判断是网络接入线路不稳定、地区性网络故障还是网站服务器针对该地区的配置有误。又如，如果在网站更新后，http 工具发现大量的 404 错误码，可能是由于页面链接更改但未正确设置重定向导致的，需要检查网站的页面管理和路由配置。

（二）及时调整系统资源或者调度行为来减少直到消除网络各类稳定性问题

1. 利用监控数据包工具

- 资源调整：
  - 根据监控数据包工具监测到的网络流量情况，如果发现某个应用程序或服务产生的网络流量过大，影响了其他关键业务的网络性能，可以对该应用的网络资源进行限制。例如，通过网络设备（如路由器或防火墙）的访问控制策略，限制该应用的最大带宽使用量，或者调整其网络优先级，确保关键业务能够获得足够的网络资源。若发现网络中存在大量非必要的网络流量（如某些广告推送或恶意软件产生的流量），可以通过设置网络过滤规则，阻止这些流量进入网络，从而释放网络带宽资源，提高网络整体性能。
- 调度优化：
  - 基于监控数据包工具提供的数据包信息，对网络流量的调度进行优化。例如，如果发现网络中存在大量实时性要求较高的数据包（如语音通话或视频会议数据包）与普通数据流量混合传输，导致实时性数据包的延迟增加，可以通过网络设备的 QoS 功能，将实时性数据包设置为高优先级，优先进行传输，确保其传输延迟满足要求。同时，根据数据包的源地址和目的地址信息，优化网络路由策略，例如，对于经常相互通信的内部网络节点，可以设置更短路径的静态路由，减少数据包在网络中的传输时间和跳数，提高网络传输效率。

2. 利用 http 工具

- 资源调整：
  - 根据 http 工具测得的 HTTP 流量数据，如果发现某个网站或服务的 HTTP 流量超出服务器的承载能力，可以考虑增加服务器资源。例如，扩展服务器的内存、CPU 或增加服务器实例数量，以提高服务器对 HTTP 请求的处理能力。若发现特定地区的用户访问网站时 HTTP 流量较大且响应较慢，可以在该地区部署内容分发网络（CDN）节点，将网站的静态资源缓存到本地节点，减少数据传输距离，提高用户访问速度，同时减轻源服务器的负载。
- 调度优化：
  - 利用 http 工具提供的信息，优化 HTTP 网络调度行为。例如，如果在监测中发现某些热门页面的 HTTP 请求量巨大，导致服务器响应延迟，可以对这些页面进行缓存策略优化，设置不同的缓存有效期和缓存级别，提高页面的重复访问效率。若发现大量的 HTTP 请求集中在某个时间段，导致服务器瞬时负载过高，可以采用负载均衡技术，将 HTTP 请求均匀分配到多个后端服务器上，避免单个服务器因过载而出现性能问题，提高整个 HTTP 服务的稳定性和可靠性。

# 七、问题和解决

## 7.1 遇到的问题

1. **工具功能实现困难**
   在实现多个基于 eBPF 的性能观测点时，遇到以下挑战：
   - 部分内核函数未公开文档，导致无法确定准确的 Hook 点。
   - 由于 eBPF 运行在沙盒环境中，某些复杂的逻辑被限制，代码设计需多次调整以满足安全性要求。
2. **性能测试过程中瓶颈显现**
   - 在高负载测试环境下，`net_watcher` 工具监测数据包处理时延时，发现内存占用超出预期，影响系统整体性能。
   - 对 `cpu_watcher` 工具进行长时间调试后，记录数据出现不完整的情况，尤其是高并发情况下的事件丢失。

## 7.2 原因分析

1. **技术实现问题**
   eBPF 技术的新颖性和复杂性对开发提出了高要求，个人在工具设计初期对 eBPF 的限制性了解不足，导致部分代码需要大幅重构。
2. **性能瓶颈**
   - 工具的 Ring Buffer 数据收集效率不足，未能满足高频数据写入需求。
   - 缺少适配高负载场景的资源分配策略。

## 7.3 解决方案

1. **技术实现优化**
   - 引入广泛使用的 eBPF 库（如 libbpf），利用其封装功能减少开发复杂度。
   - 改进代码设计，将复杂逻辑移至用户态，减少内核沙盒代码量。
2. **性能优化措施**
   - 优化 Ring Buffer 的数据读写逻辑，提高高负载情况下的数据吞吐量。
   - 针对高并发场景，设计多线程支持的环形缓冲队列，实现数据并行处理。

# 八、总结和收获

## 8.1总结

参与本次基于 eBPF 技术的 Linux 内核性能观测工具开发，我深入了解了操作系统性能监控的技术原理和实际应用。从最初的技术方案设计到后期的测试与优化，整个过程中虽然遇到了不少困难，但通过不断学习和调整，最终实现了工具的功能目标，并有效提升了系统的性能分析能力。这次经历让我更加认识到技术细节的重要性以及系统化思维的价值。

## 8.2收获

1. **技术能力的提升**
   在项目中系统学习了 eBPF 技术，并掌握了性能监控工具的设计、调试和优化方法，对 Linux 内核的运行机制有了更深入的理解。
2. **问题解决的经验**
   在面对工具开发中的兼容性和性能问题时，通过调研、测试和分析，积累了定位问题和提出解决方案的实践经验。
3. **个人成长**
   通过独立完成模块开发，锻炼了自己的沟通能力、时间管理能力以及解决复杂问题的信心。
